 // Copyright (c) 2021, Armin Biere, Johannes Kepler University Linz

 // *INDENT-OFF*
const char * usage =
"usage: generator [ <option> ... ] <name> [ <parameter> ... ]\n"
"\n"
"This is a parameterized generator for bit-level SMV models.\n"
"\n"
"The following '<option>' arguments are allowed:\n"
"\n"
"  -h | --help          print command line option summary\n"
"  -r | --reset         add an explicit 'reset' input\n"
"  -s | --stall         add an explicit 'stall' input\n"
"\n"
"  -e | --exact         exact modulo and bad bound reached checks\n"
"  -b | --exact-bad     exact bad bound reached check\n"
"  -m | --exact-modulo  exact modulo check\n"
"\n"
"Further '<name>' and the remaining arguments are as follows:\n"
"\n"
"  queue <size>                      equivalence checking of two queues\n"
"\n"
"    The 'queue' has four variants, with and without reset or enable\n"
"    inputs.  It can also be of length zero (passing the input directly\n"
"    to the output).  The exact checks options are not allowed.\n"
"\n"
"  counter <bits> <modulo> <bad>     arithmetic '<modulo>' counter\n"
"\n"
"    The 'counter' has the same four basic variations with and without\n"
"    reset or stall inputs.  It also features two variants for each of\n"
"    the two checks for reaching a certain bound, i.e., reaching the\n"
"    modulo bound (current count plus one is larger or equal to 'modulo')\n"
"    or the bad state bound (current count is larger or equal to 'bad').\n"
"    The default is to use a greater-equal comparison for both checks.\n"
"    This makes the property that the bad bound can not be reached\n"
"    inductive, i.e., 1-inductive, unless 'bound' is smaller than\n"
"    'modulo' where the property fails.  The exact modulo check makes\n"
"    the model checking problem k-inductive (k = <bad> - <modulo> + 1).\n"
"    Using a stall signal requires simple path constraints.\n"
"    In total the 'counter' has 16 variants.\n"
"\n"
"The output is written to standard output.\n"
;
// *INDENT-ON*

#include <assert.h>
#include <ctype.h>
#include <limits.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

static void
print_usage_and_exit (void)
{
  fputs (usage, stdout);
  exit (0);
}

static void die (const char *, ...) __attribute__((format (printf, 1, 2)));

static void
die (const char *fmt, ...)
{
  va_list ap;
  fputs ("generator: error: ", stderr);
  va_start (ap, fmt);
  vfprintf (stderr, fmt, ap);
  va_end (ap);
  fputc ('\n', stderr);
  exit (1);
}

static void
print_queue (const char *reset, const char *stall, unsigned size)
{
  printf ("VAR\n");
  printf ("  i : boolean;\n");
  assert (size);
  if (reset)
    printf ("  r : boolean;\n");
  if (stall)
    printf ("  s : boolean;\n");
  for (char queue = 'a'; queue <= 'b'; queue++)
    for (unsigned i = 0; i < size; i++)
      printf ("  %c%u : boolean;\n", queue, i);
  printf ("ASSIGN\n");
  for (char queue = 'a'; queue <= 'b'; queue++)
    for (unsigned i = 0; i < size; i++)
      printf ("  init (%c%u) := FALSE;\n", queue, i);
  for (char queue = 'a'; queue <= 'b'; queue++)
    for (unsigned i = 0; i < size; i++)
      {
	printf ("  next (%c%u) := ", queue, i);
	if (reset)
	  printf ("!r & ");
	if (stall)
	  printf ("((s & %c%u) | (!s & ", queue, i);
	if (i)
	  printf ("%c%u", queue, i - 1);
	else
	  printf ("i");
	if (stall)
	  printf ("))");
	printf (";\n");
      }
  printf ("SPEC\n");
  printf ("  AG (a%u = b%u)\n", size - 1, size - 1);
}

static void
print_counter_matches (unsigned bits, char name, unsigned match)
{
  for (unsigned i = 0; i < bits; i++)
    {
      const unsigned bit = (i > 31 ? 0 : (match & (1u << i)));
      const char *val = bit ? "TRUE" : "FALSE";
      printf ("  %c%u := (c%u = %s)", name, i, i, val);
      if (i)
	printf (" & %c%u", name, i - 1);
      printf (";\n");
    }
}

static void
print_counter_larger (unsigned bits, char name, unsigned limit)
{
  for (unsigned i = 0; i < bits; i++)
    {
      const unsigned bit = (i > 31 ? 0 : (limit & (1u << i)));
      if (bit)
	{
	  if (i)
	    printf ("  %c%u := c%u & %c%u;\n", name, i, i, name, i - 1);
	  else
	    printf ("  %c0 := FALSE;\n", name);
	}
      else
	{
	  if (i)
	    printf ("  %c%u := c%u | %c%u;\n", name, i, i, name, i - 1);
	  else
	    printf ("  %c0 := c0;\n", name);
	}
    }
}

static void
print_counter (const char *reset, const char *stall,
	       const char *exact_modulo, const char *exact_bad,
	       unsigned bits, unsigned modulo, unsigned bad)
{
  printf ("VAR\n");
  if (reset)
    printf ("  r : boolean;\n");
  if (stall)
    printf ("  s : boolean;\n");
  for (unsigned i = 0; i < bits; i++)
    printf ("  c%u : boolean;\n", i);
  printf ("DEFINE\n");
  printf ("  a0 := TRUE;\n");
  for (unsigned i = 1; i < bits; i++)
    printf ("  a%u := c%u & a%u;\n", i, i - 1, i - 1);
  assert (modulo);
  if (modulo == 1)
    printf ("  m%u := TRUE;\n", bits - 1);
  else if (exact_modulo)
    print_counter_matches (bits, 'm', modulo-1);
  else
    print_counter_larger (bits, 'm', modulo-2);
  if (!bad)
    printf ("  b%u := TRUE;\n", bits - 1);
  else if (exact_bad)
    print_counter_matches (bits, 'b', bad);
  else
    print_counter_larger (bits, 'b', bad - 1);
  printf ("ASSIGN\n");
  for (unsigned i = 0; i < bits; i++)
    printf ("  init (c%u) := FALSE;\n", i);
  for (unsigned i = 0; i < bits; i++)
    {
      printf ("  next (c%u) := ", i);
      if (reset)
	printf ("!r & ");
      if (stall)
	printf ("((s & c%u) | (!s & ", i);
      printf ("!m%u & (c%u != a%u)", bits - 1, i, i);
      if (stall)
	printf ("))");
      printf (";\n");
    }
  printf ("SPEC\n");
  printf ("  AG !b%u\n", bits - 1);
}

static int
is_number (const char *str, int *res_ptr)
{
  const char *p = str;
  char ch;
  if (str[0] == '0' && str[1] == '0')
    return 0;
  int res = 0;
  while (isdigit (ch = *p++))
    {
      if (INT_MAX / 10 < res)
	return 0;
      res *= 10;
      const int digit = ch - '0';
      if (INT_MAX - digit < res)
	return 0;
      res += digit;
    }
  if (ch)
    return 0;
  *res_ptr = res;
  return 1;
}

int
main (int argc, char **argv)
{
  const char *name = 0;

  const int max_parameters = 3;
  int parameters[max_parameters];
  int num_parameters = 0;

  const char *reset = 0, *stall = 0;
  const char *exact_modulo = 0, *exact_bad = 0;

  for (int i = 1; i < argc; i++)
    {
      const char *arg = argv[i];
      if (!strcmp (arg, "-h") || !strcmp (arg, "--help"))
	print_usage_and_exit ();
      else if (!strcmp (arg, "-s") || !strcmp (arg, "--stall"))
	{
	  if (stall)
	    die ("redundant options '%s' and '%s'", stall, arg);
	  stall = arg;
	}
      else if (!strcmp (arg, "-r") || !strcmp (arg, "--reset"))
	{
	  if (reset)
	    die ("redundant options '%s' and '%s'", reset, arg);
	  reset = arg;
	}
      else if (!strcmp (arg, "-m") || !strcmp (arg, "--exact-modulo"))
	{
	  if (exact_modulo)
	    die ("redundant options '%s' and '%s'", exact_modulo, arg);
	  exact_modulo = arg;
	}
      else if (!strcmp (arg, "-b") || !strcmp (arg, "--exact-bad"))
	{
	  if (exact_bad)
	    die ("redundant options '%s' and '%s'", exact_bad, arg);
	  exact_bad = arg;
	}
      else if (!strcmp (arg, "-e") || !strcmp (arg, "--exact"))
	{
	  if (exact_bad)
	    die ("redundant options '%s' and '%s'", exact_bad, arg);
	  if (exact_modulo)
	    die ("redundant options '%s' and '%s'", exact_modulo, arg);
	  exact_bad = exact_modulo = arg;
	}
      else if (arg[0] == '-')
	die ("invalid option '%s' (try '-h')", arg);
      else if (num_parameters == max_parameters)
	die ("too many parameters (try '-h')");
      else if (name)
	{
	  int tmp;
	  if (!is_number (arg, &tmp))
	    die ("invalid parameter '%s' (expected number)", arg);
	  parameters[num_parameters++] = tmp;
	}
      else
	name = arg;
    }
  if (!name)
    die ("'name' missing (try '-h')");

#define can_not_use(OPTION) \
do { \
  if (OPTION) \
    die ("can not use '%s' for '%s", OPTION, name); \
} while (0)

#define print_header() \
do { \
  printf ("MODULE main -- ./generator %s", name); \
  if (reset) \
    printf (" %s", reset); \
  if (stall) \
    printf (" %s ", stall); \
  if (exact_modulo) \
    printf (" %s", exact_modulo); \
  if (exact_bad && !(exact_modulo && !strcmp (exact_modulo, exact_bad))) \
    printf (" %s", exact_bad); \
  for (unsigned i = 0; i < num_parameters; i++) \
    printf (" %u", parameters[i]); \
  printf ("\n"); \
} while (0)

  if (!strcmp (name, "queue"))
    {
      can_not_use (exact_bad);
      can_not_use (exact_modulo);
      if (!num_parameters)
	die ("size parameter of 'queue' missing");
      if (num_parameters > 1)
	die ("more than one parameter for queue");
      const unsigned size = parameters[0];
      if (!size)
	die ("queue 'size' can not be zero");
      print_header ();
      print_queue (reset, stall, size);
    }
  else if (!strcmp (name, "counter"))
    {
      if (num_parameters != 3)
	die ("'counter' expects exactly three parameters (but got %d)",
	     num_parameters);
      const unsigned bits = parameters[0];
      const unsigned modulo = parameters[1];
      const unsigned bad = parameters[2];
      if (!bits)
	die ("number of 'bits' can not be zero");
      const unsigned bits_limit = 1u << 20;
      if (bits > bits_limit)
	die ("'counter' with too many 'bits' "
	     "(should be at most %u)", bits_limit);
      if (!modulo)
	die ("the 'modulo' parameter can not be zero");
      if (bits < 32)
	{
	  const unsigned limit = 1u << bits;
	  if (modulo > limit)
	    die ("'modulo' parameter '%u' too large (max %u = 2^%u)",
		 modulo, limit, bits);
	  if (bad >= limit)
	    die ("'bad' bound parameter' %u' too large (max %u = 2^%u-1)",
		 bad, limit - 1, bits);
	}
      print_header ();
      print_counter (reset, stall,
		     exact_modulo, exact_bad, bits, modulo, bad);
    }
  else
    die ("invalid first argument '%s'", name);

  return 0;
}
